/*
 Copyright (c) 1996 - 2004, Daniel Stenberg, <daniel@haxx.se>.
  
 All rights reserved.
  
 Permission to use, copy, modify, and distribute this software for any purpose
 with or without fee is hereby granted, provided that the above copyright
 notice and this permission notice appear in all copies.
  
 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT OF THIRD PARTY RIGHTS. IN
 NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
 DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
 OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
 OR OTHER DEALINGS IN THE SOFTWARE.
  
 Except as contained in this notice, the name of a copyright holder shall not
 be used in advertising or otherwise to promote the sale, use or other dealings
 in this Software without prior written authorization of the copyright holder. 
 */

#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <sys/types.h>

static void decodeQuantum(unsigned char *dest, const char *src)
{
	unsigned int x = 0;
	int i;
	for(i = 0; i < 4; i++) {
		if(src[i] >= 'A' && src[i] <= 'Z')
			x = (x << 6) + (unsigned int)(src[i] - 'A' + 0);
		else if(src[i] >= 'a' && src[i] <= 'z')
			x = (x << 6) + (unsigned int)(src[i] - 'a' + 26);
		else if(src[i] >= '0' && src[i] <= '9')
			x = (x << 6) + (unsigned int)(src[i] - '0' + 52);
		else if(src[i] == '+')
			x = (x << 6) + 62;
		else if(src[i] == '/')
			x = (x << 6) + 63;
		else if(src[i] == '=')
			x = (x << 6);
	}
	
	dest[2] = (unsigned char)(x & 255);
	x >>= 8;
	dest[1] = (unsigned char)(x & 255);
	x >>= 8;
	dest[0] = (unsigned char)(x & 255);
}

/*
 * Curl_base64_decode()
 *
 * Given a base64 string at src, decode it into the memory pointed to by
 * dest. Returns the length of the decoded data.
 */
size_t Curl_base64_decode(const char *src, char *dest)
{
	int length = 0;
	int equalsTerm = 0;
	int i;
	int numQuantums;
	unsigned char lastQuantum[3];
	size_t rawlen=0;
	
	while((src[length] != '=') && src[length])
		length++;
	while(src[length+equalsTerm] == '=')
		equalsTerm++;
	
	numQuantums = (length + equalsTerm) / 4;
	
	rawlen = (numQuantums * 3) - equalsTerm;
	
	for(i = 0; i < numQuantums - 1; i++) {
		decodeQuantum((unsigned char *)dest, src);
		dest += 3; src += 4;
	}
	
	decodeQuantum(lastQuantum, src);
	for(i = 0; i < 3 - equalsTerm; i++)
		dest[i] = lastQuantum[i];
	
	return rawlen;
}

/* ---- Base64 Encoding --- */
static char table64[]=
"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

/*
 * Curl_base64_encode()
 *
 * Returns the length of the newly created base64 string. The third argument
 * is a pointer to an allocated area holding the base64 data. If something
 * went wrong, -1 is returned.
 *
 */
size_t Curl_base64_encode(const char *inp, size_t insize, char **outptr)
{
	unsigned char ibuf[3];
	unsigned char obuf[4];
	int i;
	int inputparts;
	char *output;
	char *base64data;
	
	char *indata = (char *)inp;
	
	*outptr = NULL; /* set to NULL in case of failure before we reach the end */
	
	if(0 == insize)
		insize = strlen(indata);
	
	base64data = output = (char*)malloc(insize*4/3+4);
	if(NULL == output)
		return 0;
	
	while(insize > 0) {
		for (i = inputparts = 0; i < 3; i++) {
			if(insize > 0) {
				inputparts++;
				ibuf[i] = *indata;
				indata++;
				insize--;
			}
			else
				ibuf[i] = 0;
		}
		
		obuf [0] = (ibuf [0] & 0xFC) >> 2;
		obuf [1] = ((ibuf [0] & 0x03) << 4) | ((ibuf [1] & 0xF0) >> 4);
		obuf [2] = ((ibuf [1] & 0x0F) << 2) | ((ibuf [2] & 0xC0) >> 6);
		obuf [3] = ibuf [2] & 0x3F;
		
		switch(inputparts) {
			case 1: /* only one byte read */
				snprintf(output, 5, "%c%c==",
						 table64[obuf[0]],
						 table64[obuf[1]]);
				break;
			case 2: /* two bytes read */
				snprintf(output, 5, "%c%c%c=",
						 table64[obuf[0]],
						 table64[obuf[1]],
						 table64[obuf[2]]);
				break;
			default:
				snprintf(output, 5, "%c%c%c%c",
						 table64[obuf[0]],
						 table64[obuf[1]],
						 table64[obuf[2]],
						 table64[obuf[3]] );
				break;
		}
		output += 4;
	}
	*output=0;
	*outptr = base64data; /* make it return the actual data memory */
	
	return strlen(base64data); /* return the length of the new data */
}
/* ---- End of Base64 Encoding ---- */
